import renderToJavascript from '../engine-javascript/compile/render.js';
import renderToAssemblyscript from '../engine-assemblyscript/compile/render.js';
import precompile from './precompile/index.js';
import { validateSettings } from './settings.js';
import { proxyAsReadOnlyIndex } from './proxies.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
var index = (graph, nodeImplementations, target, compilationSettings) => {
    const settings = validateSettings(compilationSettings, target);
    const { precompiledCode, variableNamesIndex } = precompile({
        graph,
        nodeImplementations,
        settings,
    });
    let code;
    const renderInput = {
        precompiledCode,
        settings,
        variableNamesReadOnly: proxyAsReadOnlyIndex(variableNamesIndex),
    };
    if (target === 'javascript') {
        code = renderToJavascript(renderInput);
    }
    else if (target === 'assemblyscript') {
        code = renderToAssemblyscript(renderInput);
    }
    else {
        throw new Error(`Invalid target ${target}`);
    }
    return {
        status: 0,
        code,
    };
};

export { index as default };
