/** Asserts user provided settings are valid (or throws error) and sets default values. */
const validateSettings = (userSettings, target) => {
    const arrays = userSettings.arrays || {};
    const io = {
        messageReceivers: (userSettings.io || {}).messageReceivers || {},
        messageSenders: (userSettings.io || {}).messageSenders || {},
    };
    const debug = userSettings.debug || false;
    const audio = userSettings.audio || {
        channelCount: { in: 2, out: 2 },
        bitDepth: 64,
    };
    if (![32, 64].includes(audio.bitDepth)) {
        throw new InvalidSettingsError(`"bitDepth" can be only 32 or 64`);
    }
    const customMetadata = userSettings.customMetadata || {};
    return {
        audio,
        arrays,
        io,
        debug,
        target,
        customMetadata,
    };
};
class InvalidSettingsError extends Error {
}

export { validateSettings };
