var ASSEMBLY_SCRIPT_WASM_BINDINGS_CODE = "var AssemblyScriptWasmBindings = (function (exports) {\n    'use strict';\n\n    const _proxyGetHandlerThrowIfKeyUnknown = (target, key, path) => {\n        if (!(key in target)) {\n            if ([\n                'toJSON',\n                'Symbol(Symbol.toStringTag)',\n                'constructor',\n                '$typeof',\n                '$$typeof',\n                '@@__IMMUTABLE_ITERABLE__@@',\n                '@@__IMMUTABLE_RECORD__@@',\n                'then',\n            ].includes(key)) {\n                return true;\n            }\n            throw new Error(`namespace${path ? ` <${path.keys.join('.')}>` : ''} doesn't know key \"${String(key)}\"`);\n        }\n        return false;\n    };\n\n    const getFloatArrayType = (bitDepth) => bitDepth === 64 ? Float64Array : Float32Array;\n    const proxyAsModuleWithBindings = (rawModule, bindings) => new Proxy({}, {\n        get: (_, k) => {\n            if (bindings.hasOwnProperty(k)) {\n                const key = String(k);\n                const bindingSpec = bindings[key];\n                switch (bindingSpec.type) {\n                    case 'raw':\n                        if (k in rawModule) {\n                            return rawModule[key];\n                        }\n                        else {\n                            throw new Error(`Key ${String(key)} doesn't exist in raw module`);\n                        }\n                    case 'proxy':\n                    case 'callback':\n                        return bindingSpec.value;\n                }\n            }\n            else {\n                return undefined;\n            }\n        },\n        has: function (_, k) {\n            return k in bindings;\n        },\n        set: (_, k, newValue) => {\n            if (bindings.hasOwnProperty(String(k))) {\n                const key = String(k);\n                const bindingSpec = bindings[key];\n                if (bindingSpec.type === 'callback') {\n                    bindingSpec.value = newValue;\n                }\n                else {\n                    throw new Error(`Binding key ${String(key)} is read-only`);\n                }\n            }\n            else {\n                throw new Error(`Key ${String(k)} is not defined in bindings`);\n            }\n            return true;\n        },\n    });\n    const proxyWithEngineNameMapping = (rawModule, variableNamesIndex) => proxyWithNameMapping(rawModule, {\n        globals: variableNamesIndex.globals,\n        io: variableNamesIndex.io,\n    });\n    const proxyWithNameMapping = (rawModule, variableNamesIndex) => {\n        if (typeof variableNamesIndex === 'string') {\n            return rawModule[variableNamesIndex];\n        }\n        else if (typeof variableNamesIndex === 'object') {\n            return new Proxy(rawModule, {\n                get: (_, k) => {\n                    const key = String(k);\n                    if (key in rawModule) {\n                        return Reflect.get(rawModule, key);\n                    }\n                    else if (key in variableNamesIndex) {\n                        const nextVariableNames = variableNamesIndex[key];\n                        return proxyWithNameMapping(rawModule, nextVariableNames);\n                    }\n                    else if (_proxyGetHandlerThrowIfKeyUnknown(rawModule, key)) {\n                        return undefined;\n                    }\n                },\n                has: function (_, k) {\n                    return k in rawModule || k in variableNamesIndex;\n                },\n                set: (_, k, value) => {\n                    const key = String(k);\n                    if (key in variableNamesIndex) {\n                        const variableName = variableNamesIndex[key];\n                        if (typeof variableName !== 'string') {\n                            throw new Error(`Failed to set value for key ${String(k)}: variable name is not a string`);\n                        }\n                        return Reflect.set(rawModule, variableName, value);\n                    }\n                    else {\n                        throw new Error(`Key ${String(k)} is not defined in raw module`);\n                    }\n                },\n            });\n        }\n        else {\n            throw new Error(`Invalid name mapping`);\n        }\n    };\n\n    const liftString = (rawModule, pointer) => {\n        if (!pointer) {\n            throw new Error('Cannot lift a null pointer');\n        }\n        pointer = pointer >>> 0;\n        const end = (pointer +\n            new Uint32Array(rawModule.memory.buffer)[(pointer - 4) >>> 2]) >>>\n            1;\n        const memoryU16 = new Uint16Array(rawModule.memory.buffer);\n        let start = pointer >>> 1;\n        let string = '';\n        while (end - start > 1024) {\n            string += String.fromCharCode(...memoryU16.subarray(start, (start += 1024)));\n        }\n        return string + String.fromCharCode(...memoryU16.subarray(start, end));\n    };\n    const lowerString = (rawModule, value) => {\n        if (value == null) {\n            throw new Error('Cannot lower a null string');\n        }\n        const length = value.length, pointer = rawModule.__new(length << 1, 1) >>> 0, memoryU16 = new Uint16Array(rawModule.memory.buffer);\n        for (let i = 0; i < length; ++i)\n            memoryU16[(pointer >>> 1) + i] = value.charCodeAt(i);\n        return pointer;\n    };\n    const readTypedArray = (rawModule, constructor, pointer) => {\n        if (!pointer) {\n            throw new Error('Cannot lift a null pointer');\n        }\n        const memoryU32 = new Uint32Array(rawModule.memory.buffer);\n        return new constructor(rawModule.memory.buffer, memoryU32[(pointer + 4) >>> 2], memoryU32[(pointer + 8) >>> 2] / constructor.BYTES_PER_ELEMENT);\n    };\n    const lowerFloatArray = (rawModule, bitDepth, data) => {\n        const arrayType = getFloatArrayType(bitDepth);\n        const arrayPointer = rawModule.globals.core.createFloatArray(data.length);\n        const array = readTypedArray(rawModule, arrayType, arrayPointer);\n        array.set(data);\n        return { array, arrayPointer };\n    };\n    const lowerListOfFloatArrays = (rawModule, bitDepth, data) => {\n        const arraysPointer = rawModule.globals.core.x_createListOfArrays();\n        data.forEach((array) => {\n            const { arrayPointer } = lowerFloatArray(rawModule, bitDepth, array);\n            rawModule.globals.core.x_pushToListOfArrays(arraysPointer, arrayPointer);\n        });\n        return arraysPointer;\n    };\n    const readListOfFloatArrays = (rawModule, bitDepth, listOfArraysPointer) => {\n        const listLength = rawModule.globals.core.x_getListOfArraysLength(listOfArraysPointer);\n        const arrays = [];\n        const arrayType = getFloatArrayType(bitDepth);\n        for (let i = 0; i < listLength; i++) {\n            const arrayPointer = rawModule.globals.core.x_getListOfArraysElem(listOfArraysPointer, i);\n            arrays.push(readTypedArray(rawModule, arrayType, arrayPointer));\n        }\n        return arrays;\n    };\n\n    const instantiateWasmModule = async (wasmBuffer, wasmImports = {}) => {\n        const instanceAndModule = await WebAssembly.instantiate(wasmBuffer, {\n            env: {\n                abort: (messagePointer, _, lineNumber, columnNumber) => {\n                    const message = liftString(wasmExports, messagePointer);\n                    lineNumber = lineNumber;\n                    columnNumber = columnNumber;\n                    (() => {\n                        throw Error(`${message} at ${lineNumber}:${columnNumber}`);\n                    })();\n                },\n                seed: () => {\n                    return (() => {\n                        return Date.now() * Math.random();\n                    })();\n                },\n                'console.log': (textPointer) => {\n                    console.log(liftString(wasmExports, textPointer));\n                },\n            },\n            ...wasmImports,\n        });\n        const wasmExports = instanceAndModule.instance\n            .exports;\n        return instanceAndModule.instance;\n    };\n\n    const updateWasmInOuts = ({ refs, cache, }) => {\n        cache.wasmOutput = readTypedArray(refs.rawModule, cache.arrayType, refs.rawModule.globals.core.x_getOutput());\n        cache.wasmInput = readTypedArray(refs.rawModule, cache.arrayType, refs.rawModule.globals.core.x_getInput());\n    };\n    const createEngineLifecycleBindings = (engineContext) => {\n        const { refs, cache, metadata } = engineContext;\n        return {\n            initialize: {\n                type: 'proxy',\n                value: (sampleRate, blockSize) => {\n                    metadata.settings.audio.blockSize = blockSize;\n                    metadata.settings.audio.sampleRate = sampleRate;\n                    cache.blockSize = blockSize;\n                    refs.rawModule.initialize(sampleRate, blockSize);\n                    updateWasmInOuts(engineContext);\n                },\n            },\n            dspLoop: {\n                type: 'proxy',\n                value: (input, output) => {\n                    for (let channel = 0; channel < input.length; channel++) {\n                        cache.wasmInput.set(input[channel], channel * cache.blockSize);\n                    }\n                    updateWasmInOuts(engineContext);\n                    refs.rawModule.dspLoop();\n                    updateWasmInOuts(engineContext);\n                    for (let channel = 0; channel < output.length; channel++) {\n                        output[channel].set(cache.wasmOutput.subarray(cache.blockSize * channel, cache.blockSize * (channel + 1)));\n                    }\n                },\n            },\n        };\n    };\n\n    const createCommonsBindings = (engineContext) => {\n        const { refs, cache } = engineContext;\n        return {\n            getArray: {\n                type: 'proxy',\n                value: (arrayName) => {\n                    const arrayNamePointer = lowerString(refs.rawModule, arrayName);\n                    const arrayPointer = refs.rawModule.globals.commons.getArray(arrayNamePointer);\n                    return readTypedArray(refs.rawModule, cache.arrayType, arrayPointer);\n                },\n            },\n            setArray: {\n                type: 'proxy',\n                value: (arrayName, array) => {\n                    const stringPointer = lowerString(refs.rawModule, arrayName);\n                    const { arrayPointer } = lowerFloatArray(refs.rawModule, cache.bitDepth, array);\n                    refs.rawModule.globals.commons.setArray(stringPointer, arrayPointer);\n                    updateWasmInOuts(engineContext);\n                },\n            },\n        };\n    };\n\n    const readMetadata = async (wasmBuffer) => {\n        const inputImports = {};\n        const wasmModule = WebAssembly.Module.imports(new WebAssembly.Module(wasmBuffer));\n        wasmModule\n            .filter((imprt) => imprt.module === 'input' && imprt.kind === 'function')\n            .forEach((imprt) => (inputImports[imprt.name] = () => undefined));\n        const wasmInstance = await instantiateWasmModule(wasmBuffer, {\n            input: inputImports,\n        });\n        const rawModule = wasmInstance.exports;\n        const stringPointer = rawModule.metadata.valueOf();\n        const metadataJSON = liftString(rawModule, stringPointer);\n        return JSON.parse(metadataJSON);\n    };\n\n    const mapArray = (src, func) => {\n        const dest = {};\n        src.forEach((srcValue, i) => {\n            const [key, destValue] = func(srcValue, i);\n            dest[key] = destValue;\n        });\n        return dest;\n    };\n\n    const liftMessage = (rawModule, messagePointer) => {\n        const messageTokenTypesPointer = rawModule.globals.msg.x_getTokenTypes(messagePointer);\n        const messageTokenTypes = readTypedArray(rawModule, Int32Array, messageTokenTypesPointer);\n        const message = [];\n        messageTokenTypes.forEach((tokenType, tokenIndex) => {\n            if (tokenType === rawModule.globals.msg.FLOAT_TOKEN.valueOf()) {\n                message.push(rawModule.globals.msg.readFloatToken(messagePointer, tokenIndex));\n            }\n            else if (tokenType === rawModule.globals.msg.STRING_TOKEN.valueOf()) {\n                const stringPointer = rawModule.globals.msg.readStringToken(messagePointer, tokenIndex);\n                message.push(liftString(rawModule, stringPointer));\n            }\n        });\n        return message;\n    };\n    const lowerMessage = (rawModule, message) => {\n        const template = message.reduce((template, value) => {\n            if (typeof value === 'number') {\n                template.push(rawModule.globals.msg.FLOAT_TOKEN.valueOf());\n            }\n            else if (typeof value === 'string') {\n                template.push(rawModule.globals.msg.STRING_TOKEN.valueOf());\n                template.push(value.length);\n            }\n            else {\n                throw new Error(`invalid message value ${value}`);\n            }\n            return template;\n        }, []);\n        const templateArrayPointer = rawModule.globals.msg.x_createTemplate(template.length);\n        const loweredTemplateArray = readTypedArray(rawModule, Int32Array, templateArrayPointer);\n        loweredTemplateArray.set(template);\n        const messagePointer = rawModule.globals.msg.x_create(templateArrayPointer);\n        message.forEach((value, index) => {\n            if (typeof value === 'number') {\n                rawModule.globals.msg.writeFloatToken(messagePointer, index, value);\n            }\n            else if (typeof value === 'string') {\n                const stringPointer = lowerString(rawModule, value);\n                rawModule.globals.msg.writeStringToken(messagePointer, index, stringPointer);\n            }\n        });\n        return messagePointer;\n    };\n\n    const createIoMessageReceiversBindings = ({ metadata, refs, }) => Object.entries(metadata.settings.io.messageReceivers).reduce((bindings, [nodeId, spec]) => ({\n        ...bindings,\n        [nodeId]: {\n            type: 'proxy',\n            value: mapArray(spec, (inletId) => [\n                inletId,\n                (message) => {\n                    const messagePointer = lowerMessage(refs.rawModule, message);\n                    refs.rawModule.io.messageReceivers[nodeId][inletId](messagePointer);\n                },\n            ]),\n        },\n    }), {});\n    const createIoMessageSendersBindings = ({ metadata, }) => Object.entries(metadata.settings.io.messageSenders).reduce((bindings, [nodeId, spec]) => ({\n        ...bindings,\n        [nodeId]: {\n            type: 'proxy',\n            value: mapArray(spec, (outletId) => [\n                outletId,\n                (_) => undefined,\n            ]),\n        },\n    }), {});\n    const ioMsgSendersImports = ({ metadata, refs, }) => {\n        const wasmImports = {};\n        const { variableNamesIndex } = metadata.compilation;\n        Object.entries(metadata.settings.io.messageSenders).forEach(([nodeId, spec]) => {\n            spec.forEach((outletId) => {\n                const listenerName = variableNamesIndex.io.messageSenders[nodeId][outletId];\n                wasmImports[listenerName] = (messagePointer) => {\n                    const message = liftMessage(refs.rawModule, messagePointer);\n                    refs.engine.io.messageSenders[nodeId][outletId](message);\n                };\n            });\n        });\n        return wasmImports;\n    };\n\n    const createFsBindings = (engineContext) => {\n        const { refs, cache, metadata } = engineContext;\n        const fsExportedNames = metadata.compilation.variableNamesIndex.globals.fs;\n        return {\n            sendReadSoundFileResponse: {\n                type: 'proxy',\n                value: 'x_onReadSoundFileResponse' in fsExportedNames\n                    ? (operationId, status, sound) => {\n                        let soundPointer = 0;\n                        if (sound) {\n                            soundPointer = lowerListOfFloatArrays(refs.rawModule, cache.bitDepth, sound);\n                        }\n                        refs.rawModule.globals.fs.x_onReadSoundFileResponse(operationId, status, soundPointer);\n                        updateWasmInOuts(engineContext);\n                    }\n                    : undefined,\n            },\n            sendWriteSoundFileResponse: {\n                type: 'proxy',\n                value: 'x_onWriteSoundFileResponse' in fsExportedNames\n                    ? refs.rawModule.globals.fs.x_onWriteSoundFileResponse\n                    : undefined,\n            },\n            sendSoundStreamData: {\n                type: 'proxy',\n                value: 'x_onSoundStreamData' in fsExportedNames\n                    ? (operationId, sound) => {\n                        const soundPointer = lowerListOfFloatArrays(refs.rawModule, cache.bitDepth, sound);\n                        const writtenFrameCount = refs.rawModule.globals.fs.x_onSoundStreamData(operationId, soundPointer);\n                        updateWasmInOuts(engineContext);\n                        return writtenFrameCount;\n                    }\n                    : undefined,\n            },\n            closeSoundStream: {\n                type: 'proxy',\n                value: 'x_onCloseSoundStream' in fsExportedNames\n                    ? refs.rawModule.globals.fs.x_onCloseSoundStream\n                    : undefined,\n            },\n            onReadSoundFile: { type: 'callback', value: () => undefined },\n            onWriteSoundFile: { type: 'callback', value: () => undefined },\n            onOpenSoundReadStream: { type: 'callback', value: () => undefined },\n            onOpenSoundWriteStream: { type: 'callback', value: () => undefined },\n            onSoundStreamData: { type: 'callback', value: () => undefined },\n            onCloseSoundStream: { type: 'callback', value: () => undefined },\n        };\n    };\n    const createFsImports = (engineContext) => {\n        const wasmImports = {};\n        const { cache, metadata, refs } = engineContext;\n        const exportedNames = metadata.compilation.variableNamesIndex.globals;\n        if ('fs' in exportedNames) {\n            const nameMapping = proxyWithNameMapping(wasmImports, exportedNames.fs);\n            if ('i_readSoundFile' in exportedNames.fs) {\n                nameMapping.i_readSoundFile = (operationId, urlPointer, infoPointer) => {\n                    const url = liftString(refs.rawModule, urlPointer);\n                    const info = liftMessage(refs.rawModule, infoPointer);\n                    refs.engine.globals.fs.onReadSoundFile(operationId, url, info);\n                };\n            }\n            if ('i_writeSoundFile' in exportedNames.fs) {\n                nameMapping.i_writeSoundFile = (operationId, soundPointer, urlPointer, infoPointer) => {\n                    const sound = readListOfFloatArrays(refs.rawModule, cache.bitDepth, soundPointer);\n                    const url = liftString(refs.rawModule, urlPointer);\n                    const info = liftMessage(refs.rawModule, infoPointer);\n                    refs.engine.globals.fs.onWriteSoundFile(operationId, sound, url, info);\n                };\n            }\n            if ('i_openSoundReadStream' in exportedNames.fs) {\n                nameMapping.i_openSoundReadStream = (operationId, urlPointer, infoPointer) => {\n                    const url = liftString(refs.rawModule, urlPointer);\n                    const info = liftMessage(refs.rawModule, infoPointer);\n                    updateWasmInOuts(engineContext);\n                    refs.engine.globals.fs.onOpenSoundReadStream(operationId, url, info);\n                };\n            }\n            if ('i_openSoundWriteStream' in exportedNames.fs) {\n                nameMapping.i_openSoundWriteStream = (operationId, urlPointer, infoPointer) => {\n                    const url = liftString(refs.rawModule, urlPointer);\n                    const info = liftMessage(refs.rawModule, infoPointer);\n                    refs.engine.globals.fs.onOpenSoundWriteStream(operationId, url, info);\n                };\n            }\n            if ('i_sendSoundStreamData' in exportedNames.fs) {\n                nameMapping.i_sendSoundStreamData = (operationId, blockPointer) => {\n                    const block = readListOfFloatArrays(refs.rawModule, cache.bitDepth, blockPointer);\n                    refs.engine.globals.fs.onSoundStreamData(operationId, block);\n                };\n            }\n            if ('i_closeSoundStream' in exportedNames.fs) {\n                nameMapping.i_closeSoundStream = (...args) => refs.engine.globals.fs.onCloseSoundStream(...args);\n            }\n        }\n        return wasmImports;\n    };\n\n    const createEngine = async (wasmBuffer, additionalBindings) => {\n        const metadata = await readMetadata(wasmBuffer);\n        const bitDepth = metadata.settings.audio.bitDepth;\n        const arrayType = getFloatArrayType(bitDepth);\n        const engineContext = {\n            refs: {},\n            metadata: metadata,\n            cache: {\n                wasmOutput: new arrayType(0),\n                wasmInput: new arrayType(0),\n                arrayType,\n                bitDepth,\n                blockSize: 0,\n            },\n        };\n        const wasmImports = {\n            ...createFsImports(engineContext),\n            ...ioMsgSendersImports(engineContext),\n        };\n        const wasmInstance = await instantiateWasmModule(wasmBuffer, {\n            input: wasmImports,\n        });\n        engineContext.refs.rawModule = proxyWithEngineNameMapping(wasmInstance.exports, metadata.compilation.variableNamesIndex);\n        const engineBindings = createEngineBindings(engineContext);\n        const engine = proxyAsModuleWithBindings(engineContext.refs.rawModule, {\n            ...engineBindings,\n            ...(additionalBindings || {}),\n        });\n        engineContext.refs.engine = engine;\n        return engine;\n    };\n    const createEngineBindings = (engineContext) => {\n        const { metadata, refs } = engineContext;\n        const exportedNames = metadata.compilation.variableNamesIndex.globals;\n        const io = {\n            messageReceivers: proxyAsModuleWithBindings(refs.rawModule, createIoMessageReceiversBindings(engineContext)),\n            messageSenders: proxyAsModuleWithBindings(refs.rawModule, createIoMessageSendersBindings(engineContext)),\n        };\n        const globalsBindings = {\n            commons: {\n                type: 'proxy',\n                value: proxyAsModuleWithBindings(refs.rawModule, createCommonsBindings(engineContext)),\n            },\n        };\n        if ('fs' in exportedNames) {\n            const fs = proxyAsModuleWithBindings(refs.rawModule, createFsBindings(engineContext));\n            globalsBindings.fs = { type: 'proxy', value: fs };\n        }\n        return {\n            ...createEngineLifecycleBindings(engineContext),\n            metadata: { type: 'proxy', value: metadata },\n            globals: {\n                type: 'proxy',\n                value: proxyAsModuleWithBindings(refs.rawModule, globalsBindings),\n            },\n            io: { type: 'proxy', value: io },\n        };\n    };\n\n    exports.createEngine = createEngine;\n    exports.createEngineBindings = createEngineBindings;\n\n    return exports;\n\n})({});\n";

export { ASSEMBLY_SCRIPT_WASM_BINDINGS_CODE as default };
