var JAVA_SCRIPT_BINDINGS_CODE = "var JavaScriptBindings = (function (exports) {\n    'use strict';\n\n    const _proxyGetHandlerThrowIfKeyUnknown = (target, key, path) => {\n        if (!(key in target)) {\n            if ([\n                'toJSON',\n                'Symbol(Symbol.toStringTag)',\n                'constructor',\n                '$typeof',\n                '$$typeof',\n                '@@__IMMUTABLE_ITERABLE__@@',\n                '@@__IMMUTABLE_RECORD__@@',\n                'then',\n            ].includes(key)) {\n                return true;\n            }\n            throw new Error(`namespace${path ? ` <${path.keys.join('.')}>` : ''} doesn't know key \"${String(key)}\"`);\n        }\n        return false;\n    };\n\n    const getFloatArrayType = (bitDepth) => bitDepth === 64 ? Float64Array : Float32Array;\n    const proxyAsModuleWithBindings = (rawModule, bindings) => new Proxy({}, {\n        get: (_, k) => {\n            if (bindings.hasOwnProperty(k)) {\n                const key = String(k);\n                const bindingSpec = bindings[key];\n                switch (bindingSpec.type) {\n                    case 'raw':\n                        if (k in rawModule) {\n                            return rawModule[key];\n                        }\n                        else {\n                            throw new Error(`Key ${String(key)} doesn't exist in raw module`);\n                        }\n                    case 'proxy':\n                    case 'callback':\n                        return bindingSpec.value;\n                }\n            }\n            else {\n                return undefined;\n            }\n        },\n        has: function (_, k) {\n            return k in bindings;\n        },\n        set: (_, k, newValue) => {\n            if (bindings.hasOwnProperty(String(k))) {\n                const key = String(k);\n                const bindingSpec = bindings[key];\n                if (bindingSpec.type === 'callback') {\n                    bindingSpec.value = newValue;\n                }\n                else {\n                    throw new Error(`Binding key ${String(key)} is read-only`);\n                }\n            }\n            else {\n                throw new Error(`Key ${String(k)} is not defined in bindings`);\n            }\n            return true;\n        },\n    });\n    const proxyWithEngineNameMapping = (rawModule, variableNamesIndex) => proxyWithNameMapping(rawModule, {\n        globals: variableNamesIndex.globals,\n        io: variableNamesIndex.io,\n    });\n    const proxyWithNameMapping = (rawModule, variableNamesIndex) => {\n        if (typeof variableNamesIndex === 'string') {\n            return rawModule[variableNamesIndex];\n        }\n        else if (typeof variableNamesIndex === 'object') {\n            return new Proxy(rawModule, {\n                get: (_, k) => {\n                    const key = String(k);\n                    if (key in rawModule) {\n                        return Reflect.get(rawModule, key);\n                    }\n                    else if (key in variableNamesIndex) {\n                        const nextVariableNames = variableNamesIndex[key];\n                        return proxyWithNameMapping(rawModule, nextVariableNames);\n                    }\n                    else if (_proxyGetHandlerThrowIfKeyUnknown(rawModule, key)) {\n                        return undefined;\n                    }\n                },\n                has: function (_, k) {\n                    return k in rawModule || k in variableNamesIndex;\n                },\n                set: (_, k, value) => {\n                    const key = String(k);\n                    if (key in variableNamesIndex) {\n                        const variableName = variableNamesIndex[key];\n                        if (typeof variableName !== 'string') {\n                            throw new Error(`Failed to set value for key ${String(k)}: variable name is not a string`);\n                        }\n                        return Reflect.set(rawModule, variableName, value);\n                    }\n                    else {\n                        throw new Error(`Key ${String(k)} is not defined in raw module`);\n                    }\n                },\n            });\n        }\n        else {\n            throw new Error(`Invalid name mapping`);\n        }\n    };\n\n    const createFsModule = (rawModule) => {\n        const fsExportedNames = rawModule.metadata.compilation.variableNamesIndex.globals.fs;\n        const fs = proxyAsModuleWithBindings(rawModule, {\n            onReadSoundFile: { type: 'callback', value: () => undefined },\n            onWriteSoundFile: { type: 'callback', value: () => undefined },\n            onOpenSoundReadStream: { type: 'callback', value: () => undefined },\n            onOpenSoundWriteStream: { type: 'callback', value: () => undefined },\n            onSoundStreamData: { type: 'callback', value: () => undefined },\n            onCloseSoundStream: { type: 'callback', value: () => undefined },\n            sendReadSoundFileResponse: {\n                type: 'proxy',\n                value: 'x_onReadSoundFileResponse' in fsExportedNames\n                    ? rawModule.globals.fs.x_onReadSoundFileResponse\n                    : undefined,\n            },\n            sendWriteSoundFileResponse: {\n                type: 'proxy',\n                value: 'x_onWriteSoundFileResponse' in fsExportedNames\n                    ? rawModule.globals.fs.x_onWriteSoundFileResponse\n                    : undefined,\n            },\n            sendSoundStreamData: {\n                type: 'proxy',\n                value: 'x_onSoundStreamData' in fsExportedNames\n                    ? rawModule.globals.fs.x_onSoundStreamData\n                    : undefined,\n            },\n            closeSoundStream: {\n                type: 'proxy',\n                value: 'x_onCloseSoundStream' in fsExportedNames\n                    ? rawModule.globals.fs.x_onCloseSoundStream\n                    : undefined,\n            },\n        });\n        if ('i_openSoundWriteStream' in fsExportedNames) {\n            rawModule.globals.fs.i_openSoundWriteStream = (...args) => fs.onOpenSoundWriteStream(...args);\n        }\n        if ('i_sendSoundStreamData' in fsExportedNames) {\n            rawModule.globals.fs.i_sendSoundStreamData = (...args) => fs.onSoundStreamData(...args);\n        }\n        if ('i_openSoundReadStream' in fsExportedNames) {\n            rawModule.globals.fs.i_openSoundReadStream = (...args) => fs.onOpenSoundReadStream(...args);\n        }\n        if ('i_closeSoundStream' in fsExportedNames) {\n            rawModule.globals.fs.i_closeSoundStream = (...args) => fs.onCloseSoundStream(...args);\n        }\n        if ('i_writeSoundFile' in fsExportedNames) {\n            rawModule.globals.fs.i_writeSoundFile = (...args) => fs.onWriteSoundFile(...args);\n        }\n        if ('i_readSoundFile' in fsExportedNames) {\n            rawModule.globals.fs.i_readSoundFile = (...args) => fs.onReadSoundFile(...args);\n        }\n        return fs;\n    };\n\n    const createCommonsModule = (rawModule, metadata) => {\n        const floatArrayType = getFloatArrayType(metadata.settings.audio.bitDepth);\n        return proxyAsModuleWithBindings(rawModule, {\n            getArray: {\n                type: 'proxy',\n                value: (arrayName) => rawModule.globals.commons.getArray(arrayName),\n            },\n            setArray: {\n                type: 'proxy',\n                value: (arrayName, array) => rawModule.globals.commons.setArray(arrayName, new floatArrayType(array)),\n            },\n        });\n    };\n\n    const compileRawModule = (code) => new Function(`\n        ${code}\n        return exports\n    `)();\n    const createEngineBindings = (rawModule) => {\n        const exportedNames = rawModule.metadata.compilation.variableNamesIndex.globals;\n        const globalsBindings = {\n            commons: {\n                type: 'proxy',\n                value: createCommonsModule(rawModule, rawModule.metadata),\n            },\n        };\n        if ('fs' in exportedNames) {\n            globalsBindings.fs = { type: 'proxy', value: createFsModule(rawModule) };\n        }\n        return {\n            metadata: { type: 'raw' },\n            initialize: { type: 'raw' },\n            dspLoop: { type: 'raw' },\n            io: { type: 'raw' },\n            globals: {\n                type: 'proxy',\n                value: proxyAsModuleWithBindings(rawModule, globalsBindings),\n            },\n        };\n    };\n    const createEngine = (code, additionalBindings) => {\n        const rawModule = compileRawModule(code);\n        const rawModuleWithNameMapping = proxyWithEngineNameMapping(rawModule, rawModule.metadata.compilation.variableNamesIndex);\n        return proxyAsModuleWithBindings(rawModule, {\n            ...createEngineBindings(rawModuleWithNameMapping),\n            ...(additionalBindings || {}),\n        });\n    };\n\n    exports.compileRawModule = compileRawModule;\n    exports.createEngine = createEngine;\n    exports.createEngineBindings = createEngineBindings;\n\n    return exports;\n\n})({});\n";

export { JAVA_SCRIPT_BINDINGS_CODE as default };
